from enum import Enum


class ListEntryTypes(Enum):
    ANIME = 'anime'
    MANGA = 'manga'


class AnimeListEntryStatuses(str, Enum):
    WATCHING = 'watching'
    COMPLETED = 'completed'
    ON_HOLD = 'on_hold'
    DROPPED = 'dropped'
    PLANNED = 'planned'


class MangaListEntryStatuses(str, Enum):
    READING = 'reading'
    COMPLETED = 'completed'
    ON_HOLD = 'on_hold'
    DROPPED = 'dropped'
    PLANNED = 'plan_to_read'



from dataclasses import dataclass
from typing import Optional, Any

from pyMALv2.models.genre import Genre
from pyMALv2.models.utils import from_union, from_int, from_none, from_str


@dataclass
class Serialization:
    id: Optional[int] = None
    name: Optional[str] = None

    @staticmethod
    def from_dict(obj: Any) -> 'Serialization':
        assert isinstance(obj, dict)
        id = from_union([from_int, from_none], obj['node'].get("id"))
        name = from_union([from_str, from_none], obj['node'].get("name"))
        return Serialization(id, name)

    def to_dict(self) -> dict:
        result: dict = {}
        result["id"] = from_union([from_int, from_none], self.id)
        result["name"] = from_union([from_str, from_none], self.name)
        return result



from dataclasses import dataclass
from typing import Optional, List, Any

from pyMALv2.models.utils import from_list, from_str, from_none, from_union


@dataclass
class AlternativeTitles:
    synonyms: Optional[List[str]] = None
    en: Optional[str] = None
    ja: Optional[str] = None

    @staticmethod
    def from_dict(obj: Any) -> 'AlternativeTitles':
        assert isinstance(obj, dict)
        synonyms = from_union([lambda x: from_list(from_str, x), from_none], obj.get("synonyms"))
        en = from_union([from_str, from_none], obj.get("en"))
        ja = from_union([from_str, from_none], obj.get("ja"))
        return AlternativeTitles(synonyms, en, ja)

    def to_dict(self) -> dict:
        result: dict = {}
        result["synonyms"] = from_union([lambda x: from_list(from_str, x), from_none], self.synonyms)
        result["en"] = from_union([from_str, from_none], self.en)
        result["ja"] = from_union([from_str, from_none], self.ja)
        return result



from dataclasses import dataclass
from typing import Optional, Any

from pyMALv2.models.utils import from_str, from_none, from_union


@dataclass
class Picture:
    medium: Optional[str] = None
    large: Optional[str] = None

    @staticmethod
    def from_dict(obj: Any) -> 'Picture':
        assert isinstance(obj, dict)
        medium = from_union([from_str, from_none], obj.get("medium"))
        large = from_union([from_str, from_none], obj.get("large"))
        return Picture(medium, large)

    def to_dict(self) -> dict:
        result: dict = {}
        result["medium"] = from_union([from_str, from_none], self.medium)
        result["large"] = from_union([from_str, from_none], self.large)
        return result



from dataclasses import dataclass
from typing import Optional, Any, TYPE_CHECKING
if TYPE_CHECKING:
    from .anime import Anime
from .utils import from_str, from_none, from_union, to_class


@dataclass
class RelatedAnime:
    anime: Optional['Anime'] = None
    relation_type: Optional[str] = None
    relation_type_formatted: Optional[str] = None

    @staticmethod
    def from_dict(obj: Any) -> 'RelatedAnime':
        from .anime import Anime  # Can't import at top of file because of circular dependency
        assert isinstance(obj, dict)
        anime = from_union([Anime.from_dict, from_none], obj.get("node"))
        relation_type = from_union([from_str, from_none], obj.get("relation_type"))
        relation_type_formatted = from_union([from_str, from_none], obj.get("relation_type_formatted"))
        return RelatedAnime(anime, relation_type, relation_type_formatted)

    def to_dict(self) -> dict:
        from .anime import Anime  # Can't import at top of file because of circular dependency
        result: dict = {}
        result["anime"] = from_union([lambda x: to_class(Anime, x), from_none], self.anime)
        result["relation_type"] = from_union([from_str, from_none], self.relation_type)
        result["relation_type_formatted"] = from_union([from_str, from_none], self.relation_type_formatted)
        return result



import json
import requests

from ..exceptions import AuthorizationFailed
from ..constants.mal_endpoints import *

class Authorization():
    access_token = None
    refresh_token = None
    expires_in = None
    client_id = None
    client_secret = None

    def load_token_from_json(self, json_file):
        """
        Load token from json file.

        :param json_file: The json file containing the token.
        :return: None
        """
        with open(json_file) as f:
            data = json.load(f)
            self.access_token = data['access_token']
            self.refresh_token = data['refresh_token']
            self.expires_in = data['expires_in']
            self.client_id = data['client_id']
            self.client_secret = data['client_secret']

    def load_token(self, access_token, refresh_token, expires_in, client_id, client_secret):
        """
        Load token from parameters.

        :param client_secret: The client secret.
        :param client_id: The client id.
        :param access_token: The access token.
        :param refresh_token: The refresh token.
        :param expires_in: The token expiration time (seconds).
        :return:
        """
        self.access_token = access_token
        self.refresh_token = refresh_token
        self.expires_in = expires_in
        self.client_id = client_id
        self.client_secret = client_secret

    def check_tokens(self) -> bool:
        if self.access_token is None or self.refresh_token is None or self.expires_in is None:
            return False
        else:
            return True

    def is_valid(self) -> bool:
        """
        Check if the token is valid.

        :return: True if the token is valid, False otherwise.
        """

        if not self.check_tokens():
            return False

        r = requests.get(MAL_GET_USER_INFO_ENDPOINT(),
                         headers={'Authorization': 'Bearer ' + self.access_token})
        if r.status_code == 200:
            return True
        else:
            return False

    def is_expired(self) -> bool:
        """
        Check if the access token is expired.

        :return: True if the access token is expired, False otherwise
        """

        if not self.check_tokens():
            return False

        return self.expires_in < 60


    def refresh(self):
        """
        Refresh the access token using the refresh token.

        :return: None
        :raises: AuthorizationFailed if the refresh token is invalid
        """

        if not self.check_tokens():
            return False

        r = requests.post(MAL_TOKEN_ENDPOINT,
                          data={'grant_type': 'refresh_token', 'refresh_token': self.refresh_token})
        if r.status_code == 200:
            data = r.json()
            self.access_token = data['access_token']
            self.refresh_token = data['refresh_token']
        else:
            raise AuthorizationFailed('Failed to obtain new access token')



from dataclasses import dataclass
from typing import Optional, Any, TYPE_CHECKING
if TYPE_CHECKING:
    from .manga import Manga
from .utils import from_union, from_none, from_int, to_class


@dataclass
class MangaRecommendation:
    manga: Optional['Manga'] = None
    num_recommendations: Optional[int] = None

    @staticmethod
    def from_dict(obj: Any) -> 'MangaRecommendation':
        from .manga import Manga  # Can't import at top of file because of circular dependency
        assert isinstance(obj, dict)
        manga = from_union([Manga.from_dict, from_none], obj.get("node"))
        num_recommendations = from_union([from_int, from_none], obj.get("num_recommendations"))
        return MangaRecommendation(manga, num_recommendations)

    def to_dict(self) -> dict:
        from .manga import Manga  # Can't import at top of file because of circular dependency
        result: dict = {}
        result["manga"] = from_union([lambda x: to_class(Manga, x), from_none], self.manga)
        result["num_recommendations"] = from_union([from_int, from_none], self.num_recommendations)
        return result



from dataclasses import dataclass
from typing import Optional, Any

from pyMALv2.models.utils import from_str, from_none, from_union, is_type


@dataclass
class Status:
    watching: Optional[int] = None
    completed: Optional[int] = None
    on_hold: Optional[int] = None
    dropped: Optional[int] = None
    plan_to_watch: Optional[int] = None

    @staticmethod
    def from_dict(obj: Any) -> 'Status':
        assert isinstance(obj, dict)
        watching = from_union([from_none, lambda x: int(from_str(x))], obj.get("watching"))
        completed = from_union([from_none, lambda x: int(from_str(x))], obj.get("completed"))
        on_hold = from_union([from_none, lambda x: int(from_str(x))], obj.get("on_hold"))
        dropped = from_union([from_none, lambda x: int(from_str(x))], obj.get("dropped"))
        plan_to_watch = from_union([from_none, lambda x: int(from_str(x))], obj.get("plan_to_watch"))
        return Status(watching, completed, on_hold, dropped, plan_to_watch)

    def to_dict(self) -> dict:
        result: dict = {}
        result["watching"] = from_union([lambda x: from_none((lambda x: is_type(type(None), x))(x)), lambda x: from_str((lambda x: str((lambda x: is_type(int, x))(x)))(x))], self.watching)
        result["completed"] = from_union([lambda x: from_none((lambda x: is_type(type(None), x))(x)), lambda x: from_str((lambda x: str((lambda x: is_type(int, x))(x)))(x))], self.completed)
        result["on_hold"] = from_union([lambda x: from_none((lambda x: is_type(type(None), x))(x)), lambda x: from_str((lambda x: str((lambda x: is_type(int, x))(x)))(x))], self.on_hold)
        result["dropped"] = from_union([lambda x: from_none((lambda x: is_type(type(None), x))(x)), lambda x: from_str((lambda x: str((lambda x: is_type(int, x))(x)))(x))], self.dropped)
        result["plan_to_watch"] = from_union([lambda x: from_none((lambda x: is_type(type(None), x))(x)), lambda x: from_str((lambda x: str((lambda x: is_type(int, x))(x)))(x))], self.plan_to_watch)
        return result



from dataclasses import dataclass
from typing import Optional, Any

from pyMALv2.models.author import Author
from pyMALv2.models.utils import from_union, from_none, from_str, to_class


@dataclass
class Authors:
    author: Optional[Author] = None
    role: Optional[str] = None

    @staticmethod
    def from_dict(obj: Any) -> 'Authors':
        assert isinstance(obj, dict)
        author = from_union([Author.from_dict, from_none], obj.get("node"))
        role = from_union([from_str, from_none], obj.get("role"))
        return Authors(author, role)

    def to_dict(self) -> dict:
        result: dict = {}
        result["author"] = from_union([lambda x: to_class(Author, x), from_none], self.author)
        result["role"] = from_union([from_str, from_none], self.role)
        return result



import json

from ..base import Base
from ...auth import Authorization
from typing import List, TYPE_CHECKING
from urllib.parse import urlparse, parse_qs

from ...constants.mal_endpoints import MAL_GET_ANIME_LIST_ENDPOINT, MAL_MANGA_LIST_ENTRY_ENDPOINT, \
    MAL_GET_MANGA_LIST_ENDPOINT
from ...models.user_manga_list_entry import UserMangaListEntry
from ...models.manga import Manga
from ...models.manga_my_list_status import MangaMyListStatus


class UserMangaList(Base):
    def __init__(self, auth: Authorization):
        super().__init__(auth)
        self.entries = self._get_list()

    def entry(self, id: int):
        """
        Returns the manga entry with the given id.

        :param id: The manga id.
        :return: A UserMangaListEntry object.
        """
        for entry in self.entries:
            if entry.manga.id == id:
                return entry

    def update(self, manga: 'Manga', status: str = None, is_reading: bool = None, score: int = None,
            num_volumes_read: int = None, num_chapters_read: int = None, priority: int = None,
            num_times_reread: int = None, reread_value: int = None, tags: str = None, comments: str = None):
        """
        Adds or updates a manga entry on the user's manga list.

        :param manga: A manga object with at least an id.
        :param status: The read status of the manga.
        :param is_reading: Is currently reading the manga.
        :param score: The user's score for the manga.
        :param num_volumes_read: The number of volumes read.
        :param num_chapters_read:  The number of chapters read.
        :param priority: The priority of the manga.
        :param num_times_reread: The number of times the user has reread the manga.
        :param reread_value: The reread value of the manga.
        :param tags: Tags for this record.
        :param comments: Comments for this record.
        :return:
        """
        self._request('PATCH', MAL_MANGA_LIST_ENTRY_ENDPOINT(manga.id), data={
            'status': status,
            'is_reading': is_reading,
            'score': score,
            'num_volumes_read': num_volumes_read,
            'num_chapters_read': num_chapters_read,
            'priority': priority,
            'num_times_reread': num_times_reread,
            'reread_value': reread_value,
            'tags': tags,
            'comments': comments,
        })
        self.entries = self._get_list()

    def get(self):
        """
        Returns the user's manga list.

        :return: A list of UserMangaListEntry objects.
        """
        return self.entries

    def delete(self, id: int):
        """
        Deletes the manga entry with the given id.

        :param id: The manga id.
        :return: None
        """
        self.entry(id).delete()
        self.entries = self._get_list()

    def _get_list(self) -> List['UserMangaListEntry']:
        """
        Gets the user's manga list from MAL

        :return: A list of UserMangaListEntry objects.
        """
        offset = 0
        results = []
        while True:
            r = self._request('GET', MAL_GET_MANGA_LIST_ENDPOINT(),
                              params={'limit': 1000, 'fields': 'list_status', offset: offset})
            page = json.loads(r.text)

            for entry in page['data']:
                results.append(UserMangaListEntry.from_dict(entry, self.auth))

            if 'next' not in page['paging']:
                break
            else:
                parsed_url = urlparse(page['paging']['next'])
                next_url = parse_qs(parsed_url.query)
                offset = int(next_url['offset'][0])

        return results



from dataclasses import dataclass
from typing import Optional, Any

from pyMALv2.models.utils import from_str, from_none, from_union, from_int


@dataclass
class StartSeason:
    year: Optional[int] = None
    season: Optional[str] = None

    @staticmethod
    def from_dict(obj: Any) -> 'StartSeason':
        assert isinstance(obj, dict)
        year = from_union([from_int, from_none], obj.get("year"))
        season = from_union([from_str, from_none], obj.get("season"))
        return StartSeason(year, season)

    def to_dict(self) -> dict:
        result: dict = {}
        result["year"] = from_union([from_int, from_none], self.year)
        result["season"] = from_union([from_str, from_none], self.season)
        return result



import json

from ...auth import Authorization
from ...services.base import Base
from ...constants.mal_endpoints import *
from .user_anime_list import UserAnimeList
from .user_manga_list import UserMangaList
from types import SimpleNamespace


class UserService(Base):
    def __init__(self, auth: Authorization):
        super().__init__(auth)

    def get_current_user_info(self, fields: list = None):
        """
        Get the user's info.

        :param fields: The fields to return.
        :return: Object with the user's info.
        """
        r = self._request('GET', MAL_GET_USER_INFO_ENDPOINT(), params={'fields': ','.join(fields)})
        return json.loads(r.text, object_hook=lambda d: SimpleNamespace(**d))

    @property
    def anime_list(self):
        return UserAnimeList(self.auth)

    @property
    def manga_list(self):
        return UserMangaList(self.auth)




import json
from typing import List
from urllib.parse import urlparse, parse_qs

from pyMALv2.exceptions.MangaNotFound import MangaNotFound
from pyMALv2.models.manga import Manga
from pyMALv2.services.base import Base
from pyMALv2.auth import Authorization
from pyMALv2.constants.mal_endpoints import *

class MangaService(Base):
    def __init__(self, auth: Authorization):
        super().__init__(auth)

    def search(self, q: str, fields: str = None, limit: int = 100, offset: int = 0, get_all: bool = False) -> List[Manga]:
        """
        Searches for manga on MyAnimeList.

        :param get_all: Gets all the results.
        :param q: Search query
        :param fields: The fields to return
        :param limit: The limit of results to return.
        :param offset: The offset of the first result to return.
        :return: Object with the anime list.
        """
        results = []
        while True:
            r = self._request_no_auth('GET', MAL_SEARCH_MANGA_ENDPOINT,
                                      params={'q': q, 'fields': fields, 'limit': limit, 'offset': offset})
            page = json.loads(r.text)

            for manga in page['data']:
                results.append(Manga.from_dict(manga['node']))

            if not get_all:
                break

            if 'next' not in page['paging']:
                break
            else:
                parsed_url = urlparse(page['paging']['next'])
                next_url = parse_qs(parsed_url.query)
                offset = int(next_url['offset'][0])

        return results

    def get(self, manga_id: int, fields: str = None):
        """
        Get an manga by id.

        :param manga_id: The manga id.
        :param fields: The fields to return.
        :return: Object with the anime info.
        """
        r = self._request_no_auth('GET', MAL_GET_MANGA_INFO_ENDPOINT(manga_id), params={'fields': fields})

        if r.status_code == 404:
            raise MangaNotFound('Manga with id {} not found'.format(manga_id))
        elif r.status_code != 200:
            raise Exception('MyAnimeList returned HTTP {} with message {}'.format(r.status_code, r.json['error']))

        return Manga.from_dict(json.loads(r.text))




MAL_TOKEN_ENDPOINT = "https://myanimelist.net/v1/oauth2/token"
MAL_OAUTH_ENDPOINT = "https://myanimelist.net/v1/oauth2/authorize"
MAL_SEARCH_ANIME_ENDPOINT = "https://api.myanimelist.net/v2/anime"
MAL_SEARCH_MANGA_ENDPOINT = "https://api.myanimelist.net/v2/manga"


def MAL_GET_ANIME_INFO_ENDPOINT(anime_id: int) -> str:
    return f"https://api.myanimelist.net/v2/anime/{anime_id}"


def MAL_GET_MANGA_INFO_ENDPOINT(manga_id: int) -> str:
    return f"https://api.myanimelist.net/v2/manga/{manga_id}"


def MAL_GET_USER_INFO_ENDPOINT(username: str = '@me') -> str:
    return f"https://api.myanimelist.net/v2/users/{username}"


def MAL_GET_ANIME_LIST_ENDPOINT(username: str = '@me') -> str:
    return f"https://api.myanimelist.net/v2/users/{username}/animelist"


def MAL_GET_MANGA_LIST_ENDPOINT(username: str = '@me') -> str:
    return f"https://api.myanimelist.net/v2/users/{username}/mangalist"


def MAL_MANGA_LIST_ENTRY_ENDPOINT(manga_id: int) -> str:
    return f"https://api.myanimelist.net/v2/manga/{manga_id}/my_list_status"


def MAL_ANIME_LIST_ENTRY_ENDPOINT(anime_id: int) -> str:
    return f"https://api.myanimelist.net/v2/anime/{anime_id}/my_list_status"



from dataclasses import dataclass
from typing import Optional, List, Any
from datetime import datetime

from .alternative_titles import AlternativeTitles
from .broadcast import Broadcast
from .genre import Genre
from .anime_my_list_status import AnimeMyListStatus
from .picture import Picture
from .anime_recommendation import AnimeRecommendation
from .related_anime import RelatedAnime
from .start_season import StartSeason
from .statistics import Statistics
from .utils import from_list, from_str, from_none, from_union, from_int, from_datetime, to_class, \
    from_float, to_float


@dataclass
class Anime:
    id: Optional[int] = None
    title: Optional[str] = None
    main_picture: Optional[Picture] = None
    alternative_titles: Optional[AlternativeTitles] = None
    start_date: Optional[datetime] = None
    end_date: Optional[datetime] = None
    synopsis: Optional[str] = None
    mean: Optional[float] = None
    rank: Optional[int] = None
    popularity: Optional[int] = None
    num_list_users: Optional[int] = None
    num_scoring_users: Optional[int] = None
    nsfw: Optional[str] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    media_type: Optional[str] = None
    status: Optional[str] = None
    genres: Optional[List[Genre]] = None
    my_list_status: Optional[AnimeMyListStatus] = None
    num_episodes: Optional[int] = None
    start_season: Optional[StartSeason] = None
    broadcast: Optional[Broadcast] = None
    source: Optional[str] = None
    average_episode_duration: Optional[int] = None
    rating: Optional[str] = None
    pictures: Optional[List[Picture]] = None
    background: Optional[str] = None
    related_anime: Optional[List[RelatedAnime]] = None
    related_manga: Optional[List[Any]] = None
    recommendations: Optional[List[AnimeRecommendation]] = None
    studios: Optional[List[Genre]] = None
    statistics: Optional[Statistics] = None

    @staticmethod
    def from_dict(obj: Any) -> 'Anime':
        assert isinstance(obj, dict)
        id = from_union([from_int, from_none], obj.get("id"))
        title = from_union([from_str, from_none], obj.get("title"))
        main_picture = from_union([Picture.from_dict, from_none], obj.get("main_picture"))
        alternative_titles = from_union([AlternativeTitles.from_dict, from_none], obj.get("alternative_titles"))
        start_date = from_union([from_datetime, from_none], obj.get("start_date"))
        end_date = from_union([from_datetime, from_none], obj.get("end_date"))
        synopsis = from_union([from_str, from_none], obj.get("synopsis"))
        mean = from_union([from_float, from_none], obj.get("mean"))
        rank = from_union([from_int, from_none], obj.get("rank"))
        popularity = from_union([from_int, from_none], obj.get("popularity"))
        num_list_users = from_union([from_int, from_none], obj.get("num_list_users"))
        num_scoring_users = from_union([from_int, from_none], obj.get("num_scoring_users"))
        nsfw = from_union([from_str, from_none], obj.get("nsfw"))
        created_at = from_union([from_datetime, from_none], obj.get("created_at"))
        updated_at = from_union([from_datetime, from_none], obj.get("updated_at"))
        media_type = from_union([from_str, from_none], obj.get("media_type"))
        status = from_union([from_str, from_none], obj.get("status"))
        genres = from_union([lambda x: from_list(Genre.from_dict, x), from_none], obj.get("genres"))
        my_list_status = from_union([AnimeMyListStatus.from_dict, from_none], obj.get("my_list_status"))
        num_episodes = from_union([from_int, from_none], obj.get("num_episodes"))
        start_season = from_union([StartSeason.from_dict, from_none], obj.get("start_season"))
        broadcast = from_union([Broadcast.from_dict, from_none], obj.get("broadcast"))
        source = from_union([from_str, from_none], obj.get("source"))
        average_episode_duration = from_union([from_int, from_none], obj.get("average_episode_duration"))
        rating = from_union([from_str, from_none], obj.get("rating"))
        pictures = from_union([lambda x: from_list(Picture.from_dict, x), from_none], obj.get("pictures"))
        background = from_union([from_str, from_none], obj.get("background"))
        related_anime = from_union([lambda x: from_list(RelatedAnime.from_dict, x), from_none], obj.get("related_anime"))
        related_manga = from_union([lambda x: from_list(lambda x: x, x), from_none], obj.get("related_manga"))
        recommendations = from_union([lambda x: from_list(AnimeRecommendation.from_dict, x), from_none], obj.get("recommendations"))
        studios = from_union([lambda x: from_list(Genre.from_dict, x), from_none], obj.get("studios"))
        statistics = from_union([Statistics.from_dict, from_none], obj.get("statistics"))
        return Anime(id, title, main_picture, alternative_titles, start_date, end_date, synopsis, mean, rank, popularity, num_list_users, num_scoring_users, nsfw, created_at, updated_at, media_type, status, genres, my_list_status, num_episodes, start_season, broadcast, source, average_episode_duration, rating, pictures, background, related_anime, related_manga, recommendations, studios, statistics)

    def to_dict(self) -> dict:
        result: dict = {}
        result["id"] = from_union([from_int, from_none], self.id)
        result["title"] = from_union([from_str, from_none], self.title)
        result["main_picture"] = from_union([lambda x: to_class(Picture, x), from_none], self.main_picture)
        result["alternative_titles"] = from_union([lambda x: to_class(AlternativeTitles, x), from_none], self.alternative_titles)
        result["start_date"] = from_union([lambda x: x.isoformat(), from_none], self.start_date)
        result["end_date"] = from_union([lambda x: x.isoformat(), from_none], self.end_date)
        result["synopsis"] = from_union([from_str, from_none], self.synopsis)
        result["mean"] = from_union([to_float, from_none], self.mean)
        result["rank"] = from_union([from_int, from_none], self.rank)
        result["popularity"] = from_union([from_int, from_none], self.popularity)
        result["num_list_users"] = from_union([from_int, from_none], self.num_list_users)
        result["num_scoring_users"] = from_union([from_int, from_none], self.num_scoring_users)
        result["nsfw"] = from_union([from_str, from_none], self.nsfw)
        result["created_at"] = from_union([lambda x: x.isoformat(), from_none], self.created_at)
        result["updated_at"] = from_union([lambda x: x.isoformat(), from_none], self.updated_at)
        result["media_type"] = from_union([from_str, from_none], self.media_type)
        result["status"] = from_union([from_str, from_none], self.status)
        result["genres"] = from_union([lambda x: from_list(lambda x: to_class(Genre, x), x), from_none], self.genres)
        result["my_list_status"] = from_union([lambda x: to_class(AnimeMyListStatus, x), from_none], self.my_list_status)
        result["num_episodes"] = from_union([from_int, from_none], self.num_episodes)
        result["start_season"] = from_union([lambda x: to_class(StartSeason, x), from_none], self.start_season)
        result["broadcast"] = from_union([lambda x: to_class(Broadcast, x), from_none], self.broadcast)
        result["source"] = from_union([from_str, from_none], self.source)
        result["average_episode_duration"] = from_union([from_int, from_none], self.average_episode_duration)
        result["rating"] = from_union([from_str, from_none], self.rating)
        result["pictures"] = from_union([lambda x: from_list(lambda x: to_class(Picture, x), x), from_none], self.pictures)
        result["background"] = from_union([from_str, from_none], self.background)
        result["related_anime"] = from_union([lambda x: from_list(lambda x: to_class(RelatedAnime, x), x), from_none], self.related_anime)
        result["related_manga"] = from_union([lambda x: from_list(lambda x: x, x), from_none], self.related_manga)
        result["recommendations"] = from_union([lambda x: from_list(lambda x: to_class(AnimeRecommendation, x), x),
                                                from_none], self.recommendations)
        result["studios"] = from_union([lambda x: from_list(lambda x: to_class(Genre, x), x), from_none], self.studios)
        result["statistics"] = from_union([lambda x: to_class(Statistics, x), from_none], self.statistics)
        return result


def anime_from_dict(s: Any) -> Anime:
    return Anime.from_dict(s)


def anime_to_dict(x: Anime) -> Any:
    return to_class(Anime, x)



from dataclasses import dataclass
from typing import Optional, Any

from pyMALv2.models.utils import from_none, from_union, from_int, to_class
from pyMALv2.models.status import Status


@dataclass
class Statistics:
    status: Optional[Status] = None
    num_list_users: Optional[int] = None

    @staticmethod
    def from_dict(obj: Any) -> 'Statistics':
        assert isinstance(obj, dict)
        status = from_union([Status.from_dict, from_none], obj.get("status"))
        num_list_users = from_union([from_int, from_none], obj.get("num_list_users"))
        return Statistics(status, num_list_users)

    def to_dict(self) -> dict:
        result: dict = {}
        result["status"] = from_union([lambda x: to_class(Status, x), from_none], self.status)
        result["num_list_users"] = from_union([from_int, from_none], self.num_list_users)
        return result



import json
from typing import List, TYPE_CHECKING

from ..base import Base
from ...auth import Authorization
from ...constants.mal_endpoints import MAL_GET_ANIME_LIST_ENDPOINT, MAL_ANIME_LIST_ENTRY_ENDPOINT
from urllib.parse import urlparse, parse_qs

from ...exceptions.AnimeNotFound import AnimeNotFound
from ...models.anime_my_list_status import AnimeMyListStatus
from ...models.user_anime_list_entry import UserAnimeListEntry
from ...models.anime import Anime


class UserAnimeList(Base):
    def __init__(self, auth: Authorization):
        super().__init__(auth)
        self.entries = self._get_list()

    def get(self) -> List['UserAnimeListEntry']:
        """
        Returns a list of all anime entries on the user's anime list.

        :return: A list of anime objects.
        """
        return self.entries

    def update(self, anime: 'Anime', status: str = None, is_rewatching: bool = None, score: int = None,
               num_watched_episodes: int = None, priority: int = None, num_times_rewatched: int = None,
               rewatch_value: int = None, tags: str = None, comments: str = None):
        """
        Adds or updates an anime entry on the user's anime list.

        :param anime: An anime object with at least an id.
        :param status: The watch status of the anime.
        :param is_rewatching: Is currently rewatching the anime.
        :param score: The user's score for the anime.
        :param num_watched_episodes: The number of watched episodes.
        :param priority: The priority of the anime.
        :param num_times_rewatched: The number of times the user has rewatched the anime.
        :param rewatch_value: The rewatch value of the anime.
        :param tags: Tags for this record.
        :param comments: Comments for this record.
        :return:
        """
        self._request('PATCH', MAL_ANIME_LIST_ENTRY_ENDPOINT(anime.id), data={
            'status': status,
            'is_rewatching': is_rewatching,
            'score': score,
            'num_watched_episodes': num_watched_episodes,
            'priority': priority,
            'num_times_rewatched': num_times_rewatched,
            'rewatch_value': rewatch_value,
            'tags': tags,
            'comments': comments,
        })
        self.entries = self._get_list()

    def delete(self, id: int):
        """
        Deletes an anime entry on the user's anime list.

        :param id: The id of the anime to delete.
        :return: None
        """
        self.entry(id).delete()
        self.entries = self._get_list()

    def entry(self, id: int) -> 'UserAnimeListEntry':
        """
        Returns an anime entry on the user's anime list.

        :param id: The id of the anime to get.
        :return: A UserAnimeListEntry object.
        """
        for entry in self.entries:
            if entry.anime.id == id:
                return entry

        raise AnimeNotFound(f'Anime with id {id} not found on the user\'s anime list.')

    def _get_list(self) -> List['UserAnimeListEntry']:
        """
        Gets the user's anime list from MAL.

        :return: A list of UserAnimeListEntry objects.
        """
        offset = 0
        results = []
        while True:
            r = self._request('GET', MAL_GET_ANIME_LIST_ENDPOINT(),
                              params={'limit': 1000, 'fields': 'list_status', offset: offset})
            page = json.loads(r.text)

            for entry in page['data']:
                results.append(UserAnimeListEntry.from_dict(entry, self.auth))

            if 'next' not in page['paging']:
                break
            else:
                parsed_url = urlparse(page['paging']['next'])
                next_url = parse_qs(parsed_url.query)
                offset = int(next_url['offset'][0])

        return results



from dataclasses import dataclass
from typing import Optional, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from .anime import Anime
from .utils import from_none, from_union, from_int, to_class


@dataclass
class AnimeRecommendation:
    anime: Optional['Anime'] = None
    num_recommendations: Optional[int] = None

    @staticmethod
    def from_dict(obj: Any) -> 'AnimeRecommendation':
        from .anime import Anime  # Can't import at top of file because of circular dependency
        assert isinstance(obj, dict)
        anime = from_union([Anime.from_dict, from_none], obj.get("node"))
        num_recommendations = from_union([from_int, from_none], obj.get("num_recommendations"))
        return AnimeRecommendation(anime, num_recommendations)

    def to_dict(self) -> dict:
        from .anime import Anime  # Can't import at top of file because of circular dependency
        result: dict = {}
        result["anime"] = from_union([lambda x: to_class(Anime, x), from_none], self.anime)
        result["num_recommendations"] = from_union([from_int, from_none], self.num_recommendations)
        return result



from dataclasses import dataclass
from typing import Optional, Any

from pyMALv2.models.utils import from_str, from_none, from_union


@dataclass
class Broadcast:
    day_of_the_week: Optional[str] = None
    start_time: Optional[str] = None

    @staticmethod
    def from_dict(obj: Any) -> 'Broadcast':
        assert isinstance(obj, dict)
        day_of_the_week = from_union([from_str, from_none], obj.get("day_of_the_week"))
        start_time = from_union([from_str, from_none], obj.get("start_time"))
        return Broadcast(day_of_the_week, start_time)

    def to_dict(self) -> dict:
        result: dict = {}
        result["day_of_the_week"] = from_union([from_str, from_none], self.day_of_the_week)
        result["start_time"] = from_union([from_str, from_none], self.start_time)
        return result



# This code parses date/times, so please
#
#     pip install python-dateutil
#
# To use this code, make sure you
#
#     import json
#
# and then, to convert JSON from a string, do
#
#     result = manga_from_dict(json.loads(json_string))

from dataclasses import dataclass
from typing import Optional

from .author import Author
from .manga_my_list_status import MangaMyListStatus
from .manga_recommendation import MangaRecommendation
from .related_manga import RelatedManga
from .serialization import Serialization
from .utils import *
from .alternative_titles import AlternativeTitles
from .genre import Genre
from .picture import Picture


@dataclass
class Manga:
    id: Optional[int] = None
    title: Optional[str] = None
    main_picture: Optional[Picture] = None
    alternative_titles: Optional[AlternativeTitles] = None
    start_date: Optional[datetime] = None
    synopsis: Optional[str] = None
    mean: Optional[float] = None
    rank: Optional[int] = None
    popularity: Optional[int] = None
    num_list_users: Optional[int] = None
    num_scoring_users: Optional[int] = None
    nsfw: Optional[str] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    media_type: Optional[str] = None
    status: Optional[str] = None
    genres: Optional[List[Genre]] = None
    my_list_status: Optional[MangaMyListStatus] = None
    num_volumes: Optional[int] = None
    num_chapters: Optional[int] = None
    authors: Optional[List[Author]] = None
    pictures: Optional[List[Picture]] = None
    background: Optional[str] = None
    related_anime: Optional[List[Any]] = None
    related_manga: Optional[List[RelatedManga]] = None
    recommendations: Optional[List[MangaRecommendation]] = None
    serialization: Optional[List[Serialization]] = None

    @staticmethod
    def from_dict(obj: Any) -> 'Manga':
        assert isinstance(obj, dict)
        id = from_union([from_int, from_none], obj.get("id"))
        title = from_union([from_str, from_none], obj.get("title"))
        main_picture = from_union([Picture.from_dict, from_none], obj.get("main_picture"))
        alternative_titles = from_union([AlternativeTitles.from_dict, from_none], obj.get("alternative_titles"))
        start_date = from_union([from_datetime, from_none], obj.get("start_date"))
        synopsis = from_union([from_str, from_none], obj.get("synopsis"))
        mean = from_union([from_float, from_none], obj.get("mean"))
        rank = from_union([from_int, from_none], obj.get("rank"))
        popularity = from_union([from_int, from_none], obj.get("popularity"))
        num_list_users = from_union([from_int, from_none], obj.get("num_list_users"))
        num_scoring_users = from_union([from_int, from_none], obj.get("num_scoring_users"))
        nsfw = from_union([from_str, from_none], obj.get("nsfw"))
        created_at = from_union([from_datetime, from_none], obj.get("created_at"))
        updated_at = from_union([from_datetime, from_none], obj.get("updated_at"))
        media_type = from_union([from_str, from_none], obj.get("media_type"))
        status = from_union([from_str, from_none], obj.get("status"))
        genres = from_union([lambda x: from_list(Genre.from_dict, x), from_none], obj.get("genres"))
        my_list_status = from_union([MangaMyListStatus.from_dict, from_none], obj.get("my_list_status"))
        num_volumes = from_union([from_int, from_none], obj.get("num_volumes"))
        num_chapters = from_union([from_int, from_none], obj.get("num_chapters"))
        authors = from_union([lambda x: from_list(Author.from_dict, x), from_none], obj.get("authors"))
        pictures = from_union([lambda x: from_list(Picture.from_dict, x), from_none], obj.get("pictures"))
        background = from_union([from_str, from_none], obj.get("background"))
        related_anime = from_union([lambda x: from_list(lambda x: x, x), from_none], obj.get("related_anime"))
        related_manga = from_union([lambda x: from_list(RelatedManga.from_dict, x), from_none], obj.get("related_manga"))
        recommendations = from_union([lambda x: from_list(MangaRecommendation.from_dict, x), from_none], obj.get("recommendations"))
        serialization = from_union([lambda x: from_list(Serialization.from_dict, x), from_none], obj.get("serialization"))
        return Manga(id, title, main_picture, alternative_titles, start_date, synopsis, mean, rank, popularity, num_list_users, num_scoring_users, nsfw, created_at, updated_at, media_type, status, genres, my_list_status, num_volumes, num_chapters, authors, pictures, background, related_anime, related_manga, recommendations, serialization)

    def to_dict(self) -> dict:
        result: dict = {}
        result["id"] = from_union([from_int, from_none], self.id)
        result["title"] = from_union([from_str, from_none], self.title)
        result["main_picture"] = from_union([lambda x: to_class(Picture, x), from_none], self.main_picture)
        result["alternative_titles"] = from_union([lambda x: to_class(AlternativeTitles, x), from_none], self.alternative_titles)
        result["start_date"] = from_union([lambda x: x.isoformat(), from_none], self.start_date)
        result["synopsis"] = from_union([from_str, from_none], self.synopsis)
        result["mean"] = from_union([to_float, from_none], self.mean)
        result["rank"] = from_union([from_int, from_none], self.rank)
        result["popularity"] = from_union([from_int, from_none], self.popularity)
        result["num_list_users"] = from_union([from_int, from_none], self.num_list_users)
        result["num_scoring_users"] = from_union([from_int, from_none], self.num_scoring_users)
        result["nsfw"] = from_union([from_str, from_none], self.nsfw)
        result["created_at"] = from_union([lambda x: x.isoformat(), from_none], self.created_at)
        result["updated_at"] = from_union([lambda x: x.isoformat(), from_none], self.updated_at)
        result["media_type"] = from_union([from_str, from_none], self.media_type)
        result["status"] = from_union([from_str, from_none], self.status)
        result["genres"] = from_union([lambda x: from_list(lambda x: to_class(Genre, x), x), from_none], self.genres)
        result["my_list_status"] = from_union([lambda x: to_class(MangaMyListStatus, x), from_none], self.my_list_status)
        result["num_volumes"] = from_union([from_int, from_none], self.num_volumes)
        result["num_chapters"] = from_union([from_int, from_none], self.num_chapters)
        result["authors"] = from_union([lambda x: from_list(lambda x: to_class(Author, x), x), from_none], self.authors)
        result["pictures"] = from_union([lambda x: from_list(lambda x: to_class(Picture, x), x), from_none], self.pictures)
        result["background"] = from_union([from_str, from_none], self.background)
        result["related_anime"] = from_union([lambda x: from_list(lambda x: x, x), from_none], self.related_anime)
        result["related_manga"] = from_union([lambda x: from_list(lambda x: to_class(RelatedManga, x), x), from_none], self.related_manga)
        result["recommendations"] = from_union([lambda x: from_list(lambda x: to_class(MangaRecommendation, x), x), from_none], self.recommendations)
        result["serialization"] = from_union([lambda x: from_list(lambda x: to_class(Serialization, x), x), from_none], self.serialization)
        return result


def manga_from_dict(s: Any) -> Manga:
    return Manga.from_dict(s)


def manga_to_dict(x: Manga) -> Any:
    return to_class(Manga, x)



from datetime import datetime
from typing import TypeVar, Callable, Any, List, Type, cast

import dateutil.parser

T = TypeVar("T")


def from_list(f: Callable[[Any], T], x: Any) -> List[T]:
    assert isinstance(x, list)
    return [f(y) for y in x]


def from_str(x: Any) -> str:
    assert isinstance(x, str)
    return x


def from_none(x: Any) -> Any:
    assert x is None
    return x


def from_union(fs, x):
    for f in fs:
        try:
            return f(x)
        except:
            pass
    assert False


def from_int(x: Any) -> int:
    assert isinstance(x, int) and not isinstance(x, bool)
    return x


def from_bool(x: Any) -> bool:
    assert isinstance(x, bool)
    return x


def from_datetime(x: Any) -> datetime:
    return dateutil.parser.parse(x)


def to_class(c: Type[T], x: Any) -> dict:
    assert isinstance(x, c)
    return cast(Any, x).to_dict()


def is_type(t: Type[T], x: Any) -> T:
    assert isinstance(x, t)
    return x


def from_float(x: Any) -> float:
    assert isinstance(x, (float, int)) and not isinstance(x, bool)
    return float(x)


def to_float(x: Any) -> float:
    assert isinstance(x, float)
    return x



import json
from urllib.parse import urlparse, parse_qs
from typing import List

from ..base import Base
from ...auth import Authorization
from ...constants.mal_endpoints import *
from ...exceptions.AnimeNotFound import AnimeNotFound
from ...models.anime import Anime

class AnimeService(Base):
    def __init__(self, auth: Authorization):
        super().__init__(auth)

    def search(self, q: str, fields: str = None, limit: int = 100, offset: int = 0, get_all: bool = False) -> List[Anime]:
        """
        Searches for anime on MyAnimeList.

        :param get_all: Gets all the results.
        :param q: Search query
        :param fields: The fields to return
        :param limit: The limit of results to return.
        :param offset: The offset of the first result to return.
        :return: Object with the anime list.
        """
        results = []
        while True:
            r = self._request_no_auth('GET', MAL_SEARCH_ANIME_ENDPOINT, params={'q': q, 'fields': fields, 'limit': limit, 'offset': offset})
            page = json.loads(r.text)

            for anime in page['data']:
                results.append(Anime.from_dict(anime['node']))

            if not get_all:
                break

            if 'next' not in page['paging']:
                break
            else:
                parsed_url = urlparse(page['paging']['next'])
                next_url = parse_qs(parsed_url.query)
                offset = int(next_url['offset'][0])

        return results

    def get(self, anime_id: int, fields: str = None):
        """
        Get an anime by id.

        :param anime_id: The anime id.
        :param fields: The fields to return.
        :return: Object with the anime info.
        """
        r = self._request_no_auth('GET', MAL_GET_ANIME_INFO_ENDPOINT(anime_id), params={'fields': fields})

        if r.status_code == 404:
            raise AnimeNotFound('Anime with id {} not found'.format(anime_id))
        elif r.status_code != 200:
            raise Exception('MyAnimeList returned HTTP {} with message {}'.format(r.status_code, r.json['error']))

        return Anime.from_dict(json.loads(r.text))



import json
from dataclasses import dataclass
from typing import Optional, TYPE_CHECKING

from ..auth import Authorization
from ..constants.mal_endpoints import MAL_ANIME_LIST_ENTRY_ENDPOINT
from ..enums import AnimeListEntryStatuses
from ..services.base import Base

if TYPE_CHECKING:
    from .anime import Anime
from .utils import *
from .anime_my_list_status import AnimeMyListStatus


@dataclass
class UserAnimeListEntry(Base):
    anime: Optional['Anime'] = None
    list_status: Optional[AnimeMyListStatus] = None

    def __init__(self, anime, list_status, auth):
        super().__init__(auth)
        self.anime = anime
        self.list_status = list_status

    @staticmethod
    def from_dict(obj: Any, auth: Authorization) -> 'UserAnimeListEntry':
        from .anime import Anime  # Can't import at top of file because of circular dependency
        assert isinstance(obj, dict)
        anime = from_union([Anime.from_dict, from_none], obj.get("node"))
        list_status = from_union([AnimeMyListStatus.from_dict, from_none], obj.get("list_status"))
        return UserAnimeListEntry(anime, list_status, auth)

    def to_dict(self) -> dict:
        from .anime import Anime  # Can't import at top of file because of circular dependency
        result: dict = {}
        result["anime"] = from_union([lambda x: to_class(Anime, x), from_none], self.anime)
        result["list_status"] = from_union([lambda x: to_class(AnimeMyListStatus, x), from_none], self.list_status)
        return result

    def delete(self):
        r = self._request('DELETE', MAL_ANIME_LIST_ENTRY_ENDPOINT(self.anime.id))

        if not r.status_code == 200:
            raise Exception(f'Error deleting anime entry: {r.text}')

    def update(
            self,
            status: AnimeListEntryStatuses = None,
            is_rewatching: bool = None,
            score: int = None,
            num_watched_episodes: int = None,
            priority: int = None,
            num_times_rewatched: int = None,
            rewatch_value: int = None,
            tags: str = None,
            comments: str = None,
    ):
        """
        Update an anime list entry.
        """
        url = MAL_ANIME_LIST_ENTRY_ENDPOINT(self.anime.id)
        r = self._request('PATCH', url, data={
            'status': status,
            'is_rewatching': is_rewatching,
            'score': score,
            'num_watched_episodes': num_watched_episodes,
            'priority': priority,
            'num_times_rewatched': num_times_rewatched,
            'rewatch_value': rewatch_value,
            'tags': tags,
            'comments': comments,
        })

        if r.status_code == 200:
            self.list_status = AnimeMyListStatus.from_dict(json.loads(r.text))
            return self
        else:
            raise Exception(f'Error updating anime list entry: {r.text}')



from pathlib import Path

from setuptools import setup, find_packages

VERSION = '0.0.3'
DESCRIPTION = 'Python MyAnimeList v2 API wrapper.'

this_directory = Path(__file__).parent
LONG_DESCRIPTION = (this_directory / 'README.md').read_text()

setup(
    name='pyMALv2',
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author="Clarence",
    author_email="public@cm0.dev",
    packages=find_packages(),
    install_requires=[],
    license='MIT',
    keywords=['python', 'myanimelist', 'api', 'wrapper', 'v2'],
    url='https://github.com/mclarence/pyMALv2',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],

)


import json
import secrets
import urllib.parse

from http.server import HTTPServer, BaseHTTPRequestHandler

import requests

from ..exceptions import InvalidAuthorizationCodeError, AuthorizationFailed
from ..constants.mal_endpoints import *


def _get_new_code_verifier() -> str:
    token = secrets.token_urlsafe(100)
    return token[:128]


class RequestHandler(BaseHTTPRequestHandler):

    def __init__(self):
        self.auth_code = None

    def __call__(self, *args, **kwargs):
        """Handle a request."""
        super().__init__(*args, **kwargs)

    def do_GET(self):
        self.auth_code = urllib.parse.parse_qs(urllib.parse.urlparse(self.path).query)['code'][0]


class OAuth():

    def __init__(self, client_id: str = None, client_secret: str = None, redirect_uri: str = None):
        self.client_id = client_id
        self.client_secret = client_secret
        self.redirect_uri = redirect_uri

    def start_oauth2_flow(self, http_server: bool = False, port: int = 8989):
        """
        Start the OAuth2 flow.

        :param http_server: If True, start a local HTTP server to receive the authorization code.
        :param port: The port to use for the HTTP server.
        :return:
        """
        code_verifier = _get_new_code_verifier()
        if http_server:
            authorization_code = self._receive_oauth_token(code_verifier, port)
        else:
            authorization_code = self._input_oauth_token(code_verifier)

        self.tokens = self._get_user_tokens(authorization_code, code_verifier)

    def _input_oauth_token(self, code_verifier: str):
        """
        Get the authorization code from the user.

        :param code_verifier: The code verifier.
        :return: The authorization code.
        """
        print('Please open the following URL in your browser:' + self._get_oauth2_url(code_verifier))
        return input('Please visit the URL and paste the authorization code here:')

    def _receive_oauth_token(self, code_verifier: str, port: int = 8989):
        """
        Use a local HTTP server to receive the authorization code.

        :param code_verifier: The code verifier.
        :param port: The HTTP server port.
        :return: The authorization code.
        """
        request_handler = RequestHandler()
        httpd = HTTPServer(('localhost', port), request_handler)
        print("Please open the following URL in your browser:" + self._get_oauth2_url(code_verifier))
        print("Waiting for code...")
        httpd.handle_request()
        return request_handler.auth_code

    def tokens_as_json(self):
        """
        Get the tokens as a JSON string.

        :return: The tokens as a JSON string.
        """
        # combine client_id and client_secret and tokens into one dict
        return json.dumps({**self.tokens, 'client_id': self.client_id, 'client_secret': self.client_secret})

    def _get_user_tokens(self, authorization_code: str, code_verifier: str):
        """
        Use the authorization code to get the user tokens.

        :param authorization_code: The authorization code.
        :param code_verifier: The code verifier.
        :return: The user tokens as json dict.
        """
        body = {
            'grant_type': 'authorization_code',
            'code': authorization_code,
            'code_verifier': code_verifier,
            'client_id': self.client_id,
            'client_secret': self.client_secret,
        }

        if self.redirect_uri is not None:
            body['redirect_uri'] = self.redirect_uri

        r = requests.post(MAL_TOKEN_ENDPOINT, data=body)
        if r.status_code == 200:
            return r.json()
        elif r.status_code == 401 or r.status_code == 403:
            raise InvalidAuthorizationCodeError(r.text)
        else:
            raise AuthorizationFailed(r.text)

    def _get_oauth2_url(self, code_challenge: str):
        params = {
            'response_type': 'code',
            'client_id': self.client_id,
            'code_challenge': code_challenge,
        }

        if self.redirect_uri is not None:
            params['redirect_uri'] = self.redirect_uri

        return MAL_OAUTH_ENDPOINT + '?' + urllib.parse.urlencode(params)

    def from_json_file(self, file_path: str):
        with open(file_path, 'r') as f:
            credentials = json.load(f)
            self.client_id = credentials['client_id']
            self.client_secret = credentials['client_secret']
            self.redirect_uri = credentials['redirect_uri']




from ..auth import Authorization
import requests

class Base():
    def __init__(self, auth: Authorization):
        self.auth = auth

    def _request(self, method: str, endpoint: str, params: dict = None, data: dict = None):
        """
        Make a request to the API.
        """
        if not self.auth.is_valid():
            self.auth.refresh()
        headers = {
            'Authorization': 'Bearer {}'.format(self.auth.access_token),
            'Content-Type': 'application/x-www-form-urlencoded'
        }
        response = requests.request(method, endpoint, headers=headers, params=params, data=data)
        if response.status_code == 401:
            self.auth.refresh()
            headers = {
                'Authorization': 'Bearer {}'.format(self.auth.access_token),
                'Content-Type': 'application/x-www-form-urlencoded'
            }
            response = requests.request(method, endpoint, headers=headers, params=params, data=data)
        return response

    def _request_no_auth(self, method: str, endpoint: str, params: dict = None, data: dict = None):
        headers = {
            'X-MAL-CLIENT-ID': '{}'.format(self.auth.client_id),
            'Content-Type': 'application/x-www-form-urlencoded'
        }
        return requests.request(method, endpoint, headers=headers, params=params, data=data)






from dataclasses import dataclass
from typing import Optional, Any

from pyMALv2.models.utils import from_union, from_int, from_none, from_str


@dataclass
class Author:
    id: Optional[int] = None
    first_name: Optional[str] = None
    last_name: Optional[str] = None

    @staticmethod
    def from_dict(obj: Any) -> 'Author':
        assert isinstance(obj, dict)
        id = from_union([from_int, from_none], obj.get("id"))
        first_name = from_union([from_str, from_none], obj.get("first_name"))
        last_name = from_union([from_str, from_none], obj.get("last_name"))
        return Author(id, first_name, last_name)

    def to_dict(self) -> dict:
        result: dict = {}
        result["id"] = from_union([from_int, from_none], self.id)
        result["first_name"] = from_union([from_str, from_none], self.first_name)
        result["last_name"] = from_union([from_str, from_none], self.last_name)
        return result



from dataclasses import dataclass
from datetime import datetime
from typing import Optional, Any

from pyMALv2.models.utils import from_str, from_none, from_union, from_int, from_bool, from_datetime


@dataclass
class AnimeMyListStatus:
    status: Optional[str] = None
    score: Optional[int] = None
    num_episodes_watched: Optional[int] = None
    is_rewatching: Optional[bool] = None
    updated_at: Optional[datetime] = None

    @staticmethod
    def from_dict(obj: Any) -> 'AnimeMyListStatus':
        assert isinstance(obj, dict)
        status = from_union([from_str, from_none], obj.get("status"))
        score = from_union([from_int, from_none], obj.get("score"))
        num_episodes_watched = from_union([from_int, from_none], obj.get("num_episodes_watched"))
        is_rewatching = from_union([from_bool, from_none], obj.get("is_rewatching"))
        updated_at = from_union([from_datetime, from_none], obj.get("updated_at"))
        return AnimeMyListStatus(status, score, num_episodes_watched, is_rewatching, updated_at)

    def to_dict(self) -> dict:
        result: dict = {}
        result["status"] = from_union([from_str, from_none], self.status)
        result["score"] = from_union([from_int, from_none], self.score)
        result["num_episodes_watched"] = from_union([from_int, from_none], self.num_episodes_watched)
        result["is_rewatching"] = from_union([from_bool, from_none], self.is_rewatching)
        result["updated_at"] = from_union([lambda x: x.isoformat(), from_none], self.updated_at)
        return result



from dataclasses import dataclass
from datetime import datetime
from typing import Optional, Any

from pyMALv2.models.utils import from_union, from_str, from_none, from_bool, from_int, from_datetime


@dataclass
class MangaMyListStatus:
    status: Optional[str] = None
    is_rereading: Optional[bool] = None
    num_volumes_read: Optional[int] = None
    num_chapters_read: Optional[int] = None
    score: Optional[int] = None
    updated_at: Optional[datetime] = None

    @staticmethod
    def from_dict(obj: Any) -> 'MangaMyListStatus':
        assert isinstance(obj, dict)
        status = from_union([from_str, from_none], obj.get("status"))
        is_rereading = from_union([from_bool, from_none], obj.get("is_rereading"))
        num_volumes_read = from_union([from_int, from_none], obj.get("num_volumes_read"))
        num_chapters_read = from_union([from_int, from_none], obj.get("num_chapters_read"))
        score = from_union([from_int, from_none], obj.get("score"))
        updated_at = from_union([from_datetime, from_none], obj.get("updated_at"))
        return MangaMyListStatus(status, is_rereading, num_volumes_read, num_chapters_read, score, updated_at)

    def to_dict(self) -> dict:
        result: dict = {}
        result["status"] = from_union([from_str, from_none], self.status)
        result["is_rereading"] = from_union([from_bool, from_none], self.is_rereading)
        result["num_volumes_read"] = from_union([from_int, from_none], self.num_volumes_read)
        result["num_chapters_read"] = from_union([from_int, from_none], self.num_chapters_read)
        result["score"] = from_union([from_int, from_none], self.score)
        result["updated_at"] = from_union([lambda x: x.isoformat(), from_none], self.updated_at)
        return result
